"""
生活保護の実装
"""

import csv
from functools import cache
import json

import numpy as np
from openfisca_core.indexed_enums import Enum
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan import COUNTRY_DIR
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


# NOTE: 各種基準額表は項目数が多いため可読性の高いCSV形式やjson形式としている。
# https://www.mhlw.go.jp/content/000776372.pdf を参照


@cache
def 生活扶助基準1_第1類_基準額1表():
    """
    csvファイルから値を読み込み

    生活扶助基準1_第1類_基準額1表()[年齢, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/生活扶助基準額/第1類1.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 生活扶助基準1_逓減率1表():
    """
    csvファイルから値を読み込み

    生活扶助基準1_逓減率1表()[世帯人数, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/生活扶助基準額/逓減率1.csv",
                         delimiter=",", skip_header=1, dtype="float64")[:, 1:]


@cache
def 生活扶助基準1_第2類_基準額1表():
    """
    csvファイルから値を読み込み

    生活扶助基準1_第2類_基準額1表()[世帯人数, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/生活扶助基準額/第2類1.csv",
                        delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 生活扶助基準2_第1類_基準額2表():
    """
    csvファイルから値を読み込み

    生活扶助基準2_第1類_基準額2表()[年齢, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/生活扶助基準額/第1類2.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 生活扶助基準2_逓減率2表():
    """
    csvファイルから値を読み込み

    生活扶助基準2_逓減率2表()[世帯人数, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/生活扶助基準額/逓減率2.csv",
                         delimiter=",", skip_header=1, dtype="float64")[:, 1:]


@cache
def 生活扶助基準2_第2類_基準額2表():
    """
    csvファイルから値を読み込み

    生活扶助基準2_第2類_基準額2表()[世帯人数, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/生活扶助基準額/第2類2.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 地域区分表():
    """
    jsonファイルから値を読み込み

    地域区分表()[都道府県] の形で参照可能
    票に含まれていないものはすべて6区
    """
    with open(COUNTRY_DIR + "/assets/福祉/生活保護/冬季加算/地域区分.json") as f:
        d = json.load(f)
        return np.array(list(d.values()))


@cache
def 地域区分表_キー一覧():
    """
    jsonファイルから値を読み込み

    selectする際のキー一覧として都道府県名を取得
    """
    with open(COUNTRY_DIR + "/assets/福祉/生活保護/冬季加算/地域区分.json") as f:
        d = json.load(f)
        return d.keys()


@cache
def 冬季加算表():
    """
    csvファイルから値を読み込み

    冬季加算表()[冬季加算地域区分, 世帯人数, 居住級地区分] の形で参照可能
    """
    冬季加算表 = []
    for i in range(1, 7):
        地域区分の冬季加算表 = np.genfromtxt(COUNTRY_DIR + f"/assets/福祉/生活保護/冬季加算/{i}区.csv",
                                   delimiter=",", skip_header=1, dtype="int64")[:, 1:]
        冬季加算表.append(地域区分の冬季加算表)
    return np.stack(冬季加算表)


@cache
def 市ごとの住宅扶助限度額():
    """
    csvファイルから値を読み込み

    市ごとの住宅扶助限度額()[市, 世帯人員] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/住宅扶助基準額/市.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 市ごとの住宅扶助限度額_キー一覧():
    """
    csvファイルから値を読み込み

    selectする際のキー一覧として市名を取得
    """
    with open(COUNTRY_DIR + "/assets/福祉/生活保護/住宅扶助基準額/市.csv") as f:
        reader = csv.DictReader(f)
        return [row["市"] for row in reader]


@cache
def 都道府県ごとの住宅扶助限度額():
    """
    csvファイルから値を読み込み

    都道府県ごとの住宅扶助限度額()[都道府県区分, 世帯人員] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/住宅扶助基準額/都道府県.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 2:]


@cache
def 都道府県ごとの住宅扶助限度額_キー一覧():
    """
    csvファイルから値を読み込み

    selectする際のキー一覧として都道府県名、級地区分を取得
    """
    with open(COUNTRY_DIR + "/assets/福祉/生活保護/住宅扶助基準額/都道府県.csv") as f:
        reader = csv.DictReader(f)
        return [{"都道府県": row["都道府県"], "級地": int(row["級地"])} for row in reader]


@cache
def 生活扶助本体に係る経過的加算表():
    """
    csvファイルから値を読み込み

    生活扶助本体に係る経過的加算表()[世帯人数, 年齢, 居住級地区分1] の形で参照可能
    """
    生活扶助本体に係る経過的加算表 = []
    for i in range(1, 6):
        生活扶助本体に係る経過的加算表.append(np.genfromtxt(COUNTRY_DIR + f"/assets/福祉/生活保護/生活扶助本体に係る経過的加算/{i}人世帯.csv",
                                         delimiter=",", skip_header=1, dtype="int64")[:, 1:])
    return np.stack(生活扶助本体に係る経過的加算表)


@cache
def 母子世帯等に係る経過的加算表():
    """
    csvファイルから値を読み込み

    母子世帯等に係る経過的加算表()[世帯人数, 年齢, 居住級地区分] の形で参照可能
    """
    母子世帯等に係る経過的加算表 = []

    母子世帯等に係る経過的加算表.append(np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/母子世帯等に係る経過的加算/3人世帯.csv",
                                     delimiter=",", skip_header=1, dtype="int64")[:, 1:])
    母子世帯等に係る経過的加算表.append(np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/母子世帯等に係る経過的加算/4人世帯.csv",
                                     delimiter=",", skip_header=1, dtype="int64")[:, 1:])
    母子世帯等に係る経過的加算表.append(np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/母子世帯等に係る経過的加算/5人世帯.csv",
                                     delimiter=",", skip_header=1, dtype="int64")[:, 1:])

    return np.stack(母子世帯等に係る経過的加算表)


@cache
def 障害者加算表():
    """
    csvファイルから値を読み込み

    障害者加算表()[等級, 居住級地区分1] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/障害者加算.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


@cache
def 期末一時扶助表():
    """
    csvファイルから値を読み込み

    期末一時扶助表()[世帯人数, 居住級地区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/福祉/生活保護/期末一時扶助.csv",
                         delimiter=",", skip_header=1, dtype="int64")[:, 1:]


class 生活保護(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活保護"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        # 以下、必要なvariableを作成する。
        # 細かく作成した方が単体テストの数が少なくなるため楽。
        # 組み合わせたvariableは条件も組み合わせてテストするためテスト数が多くなる。

        # 【Ａ】 (「生活扶助基準（第１類＋第２類）①×0.855」又は「生活扶助基準（第１類＋第２類）②」のいずれか高い方)＋生活扶助本体に係る経過的加算
        生活扶助基準1 = 対象世帯("生活扶助基準1", 対象期間)
        生活扶助基準2 = 対象世帯("生活扶助基準2", 対象期間)
        生活扶助本体に係る経過的加算 = 対象世帯("生活扶助本体に係る経過的加算", 対象期間)
        # (noqa) np.maxを世帯員抽出ではなく2つの式の比較に使用しているためlinterを許容
        # NOTE: 世帯ごとの最大値を取得するため、axisの指定が必要（指定しないとすべての要素の最大値を返してしまう）
        a = np.max([生活扶助基準1 * 0.855, 生活扶助基準2], axis=0) + 生活扶助本体に係る経過的加算  # noqa: TID251

        # 【Ｂ】加算
        障害者加算 = 対象世帯("障害者加算", 対象期間)
        母子加算 = 対象世帯("母子加算", 対象期間)
        児童を養育する場合の加算 = 対象世帯("児童を養育する場合の加算", 対象期間)
        母子世帯等に係る経過的加算 = 対象世帯("母子世帯等に係る経過的加算", 対象期間)
        児童を養育する場合に係る経過的加算 = 対象世帯("児童を養育する場合に係る経過的加算", 対象期間)
        放射線障害者加算 = 対象世帯("放射線障害者加算", 対象期間)
        妊産婦加算 = 対象世帯("妊産婦加算", 対象期間)
        介護施設入所者加算 = 対象世帯("介護施設入所者加算", 対象期間)
        在宅患者加算 = 対象世帯("在宅患者加算", 対象期間)

        # 障害者加算と母子加算は併給できない（参考：https://www.mhlw.go.jp/content/000776372.pdf）
        # 高い方のみ加算（参考：https://www.ace-room.jp/safetynet/safetyqa/safety-add/）
        # (noqa) np.maxを世帯員抽出ではなく2つの式の比較に使用しているためlinterを許容
        # NOTE: 世帯ごとの最大値を取得するため、axisの指定が必要（指定しないとすべての要素の最大値を返してしまう）
        b = np.max([障害者加算, 母子加算], axis=0) + 児童を養育する場合の加算 + 母子世帯等に係る経過的加算 + 児童を養育する場合に係る経過的加算 + \
            放射線障害者加算 + 妊産婦加算 + 介護施設入所者加算 + 在宅患者加算  # noqa: TID251

        # 【Ｃ】住宅扶助基準
        住宅扶助基準 = 対象世帯("住宅扶助基準", 対象期間)
        c = 住宅扶助基準

        # 【Ｄ】教育扶助基準、高等学校等就学費
        教育扶助基準 = 対象世帯("教育扶助基準", 対象期間)
        高等学校等就学費 = 対象世帯("高等学校等就学費", 対象期間)
        d = 教育扶助基準 + 高等学校等就学費

        # NOTE: 介護費・医療費等その他の加算・扶助は実費のため計算せず、計算結果GUIの説明欄に記載
        # 【Ｅ】介護扶助基準
        e = 0.0
        # 【Ｆ】医療扶助基準
        f = 0.0

        期末一時扶助 = 対象世帯("期末一時扶助", 対象期間)

        # 以上のステップで出した最低生活費から月収と各種手当額を引いた額が給付される
        # 参考: https://www.mhlw.go.jp/stf/seisakunitsuite/bunya/hukushi_kaigo/seikatsuhogo/seikatuhogo/index.html
        最低生活費 = a + b + c + d + e + f + 期末一時扶助
        収入 = 対象世帯.sum(対象世帯.members("収入", 対象期間))
        # 就労収入のうち一定額を控除
        勤労控除 = 対象世帯("勤労控除", 対象期間)
        月収 = np.clip(収入 / 12 - 勤労控除, 0, None)

        # TODO: 実装される手当てが増えるたびに追記しなくてもよい仕組みが必要？
        各種手当額 = 対象世帯("児童手当", 対象期間) + 対象世帯("児童育成手当", 対象期間) + 対象世帯("児童扶養手当_最小", 対象期間)

        # 負の数にならないよう、0円未満になった場合は0円に補正
        return np.clip(最低生活費 - 月収 - 各種手当額, 0.0, None)


class 生活扶助基準1(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準（第1類+第2類）①"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        生活扶助基準1_第1類_基準額1 = 対象世帯("生活扶助基準1_第1類_基準額1", 対象期間)
        生活扶助基準1_逓減率1 = 対象世帯("生活扶助基準1_逓減率1", 対象期間)
        生活扶助基準1_第2類_基準額1 = 対象世帯("生活扶助基準1_第2類_基準額1", 対象期間)
        冬季加算 = 対象世帯("冬季加算", 対象期間)

        return 生活扶助基準1_第1類_基準額1 * 生活扶助基準1_逓減率1 + 生活扶助基準1_第2類_基準額1 + 冬季加算


class 生活扶助基準1_第1類_基準額1(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準(第1類) 基準額1"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        # NOTE: 世帯で計算すると配列のサイズが合わずエラーになるため、人物ごとに計算
        各世帯員の基準額 = 対象世帯.members("生活扶助基準1_第1類_基準額1_世帯員", 対象期間)
        return 対象世帯.sum(各世帯員の基準額)


class 生活扶助基準1_第1類_基準額1_世帯員(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "生活扶助基準(第1類) 基準額1"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象人物, 対象期間, _parameters):
        居住級地区分1 = 対象人物.世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象人物.世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        年齢 = 対象人物("年齢", 対象期間)
        年齢区分インデックス = np.select(
            [(年齢 >= 0) * (年齢 <= 2),
             (年齢 >= 3) * (年齢 <= 5),
             (年齢 >= 6) * (年齢 <= 11),
             (年齢 >= 12) * (年齢 <= 17),
             (年齢 >= 18) * (年齢 <= 19),
             (年齢 >= 20) * (年齢 <= 40),
             (年齢 >= 41) * (年齢 <= 59),
             (年齢 >= 60) * (年齢 <= 64),
             (年齢 >= 65) * (年齢 <= 69),
             (年齢 >= 70) * (年齢 <= 74),
             年齢 >= 75],
            list(range(11)),
            -1).astype(int)

        # NOTE: この行はentityを人物にして計算する必要がある
        # (世帯の場合、年齢区分の要素数が人物、居住級地区分の要素数が世帯となり配列の形式が一致せずエラーになる)
        各世帯員の基準額 = 生活扶助基準1_第1類_基準額1表()[年齢区分インデックス, 居住級地区分インデックス]

        return 各世帯員の基準額


class 生活扶助基準1_逓減率1(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準 逓減率1"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        # 入院患者、施設入所者は世帯人数に含めない
        入院中 = 対象世帯.members("入院中", 対象期間)
        介護施設入所中 = 対象世帯.members("介護施設入所中", 対象期間)
        世帯人数 = 対象世帯.sum(np.logical_not(入院中) * np.logical_not(介護施設入所中))
        # インデックスは0始まりなので1を引いて調整
        世帯人数区分インデックス = np.clip(世帯人数 - 1, 0, 4).astype(int)

        return 生活扶助基準1_逓減率1表()[世帯人数区分インデックス, 居住級地区分インデックス]


class 生活扶助基準1_第2類_基準額1(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準(第2類) 基準額1"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        # 入院患者、施設入所者は世帯人数に含めない
        入院中 = 対象世帯.members("入院中", 対象期間)
        介護施設入所中 = 対象世帯.members("介護施設入所中", 対象期間)
        世帯人数 = 対象世帯.sum(np.logical_not(入院中) * np.logical_not(介護施設入所中))
        # インデックスは0始まりなので1を引いて調整
        世帯人数区分インデックス = np.clip(世帯人数 - 1, 0, 4).astype(int)

        return 生活扶助基準1_第2類_基準額1表()[世帯人数区分インデックス, 居住級地区分インデックス]


class 生活扶助基準2(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準（第1類+第2類）②"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        生活扶助基準2_第1類_基準額2 = 対象世帯("生活扶助基準2_第1類_基準額2", 対象期間)
        生活扶助基準2_逓減率2 = 対象世帯("生活扶助基準2_逓減率2", 対象期間)
        生活扶助基準2_第2類_基準額2 = 対象世帯("生活扶助基準2_第2類_基準額2", 対象期間)
        冬季加算 = 対象世帯("冬季加算", 対象期間)

        return 生活扶助基準2_第1類_基準額2 * 生活扶助基準2_逓減率2 + 生活扶助基準2_第2類_基準額2 + 冬季加算


class 生活扶助基準2_第1類_基準額2(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準(第1類) 基準額2"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        各世帯員の基準額 = 対象世帯.members("生活扶助基準2_第1類_基準額2_世帯員", 対象期間)
        return 対象世帯.sum(各世帯員の基準額)


class 生活扶助基準2_第1類_基準額2_世帯員(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "生活扶助基準(第1類) 基準額2"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象人物, 対象期間, parameters):
        居住級地区分1 = 対象人物.世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象人物.世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        年齢 = 対象人物("年齢", 対象期間)
        年齢区分インデックス = np.select(
            [(年齢 >= 0) * (年齢 <= 2),
             (年齢 >= 3) * (年齢 <= 5),
             (年齢 >= 6) * (年齢 <= 11),
             (年齢 >= 12) * (年齢 <= 17),
             (年齢 >= 18) * (年齢 <= 19),
             (年齢 >= 20) * (年齢 <= 40),
             (年齢 >= 41) * (年齢 <= 59),
             (年齢 >= 60) * (年齢 <= 64),
             (年齢 >= 65) * (年齢 <= 69),
             (年齢 >= 70) * (年齢 <= 74),
             年齢 >= 75],
            list(range(11)),
            -1).astype(int)

        # NOTE: この行はentityを人物にして計算する必要がある
        # (世帯の場合、年齢区分の要素数が人物、居住級地区分の要素数が世帯となり配列の形式が一致せずエラーになる)
        各世帯員の基準額 = 生活扶助基準2_第1類_基準額2表()[年齢区分インデックス, 居住級地区分インデックス]

        return 各世帯員の基準額


class 生活扶助基準2_逓減率2(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準 逓減率2"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        # 入院患者、施設入所者は世帯人数に含めない
        入院中 = 対象世帯.members("入院中", 対象期間)
        介護施設入所中 = 対象世帯.members("介護施設入所中", 対象期間)
        世帯人数 = 対象世帯.sum(np.logical_not(入院中) * np.logical_not(介護施設入所中))
        # インデックスは0始まりなので1を引いて調整
        世帯人数区分インデックス = np.clip(世帯人数 - 1, 0, 4).astype(int)

        return 生活扶助基準2_逓減率2表()[世帯人数区分インデックス, 居住級地区分インデックス]


class 生活扶助基準2_第2類_基準額2(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助基準(第2類) 基準額2"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        # 入院患者、施設入所者は世帯人数に含めない
        入院中 = 対象世帯.members("入院中", 対象期間)
        介護施設入所中 = 対象世帯.members("介護施設入所中", 対象期間)
        世帯人数 = 対象世帯.sum(np.logical_not(入院中) * np.logical_not(介護施設入所中))
        # インデックスは0始まりなので1を引いて調整
        世帯人数区分インデックス = np.clip(世帯人数 - 1, 0, 4).astype(int)

        return 生活扶助基準2_第2類_基準額2表()[世帯人数区分インデックス, 居住級地区分インデックス]


class 生活扶助本体に係る経過的加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活扶助本体に係る経過的加算"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        # NOTE: 世帯で計算すると配列のサイズが合わずエラーになるため、人物ごとに計算
        加算額 = 対象世帯.members("生活扶助本体に係る経過的加算_世帯員", 対象期間)
        return 対象世帯.sum(加算額)


class 生活扶助本体に係る経過的加算_世帯員(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "生活扶助本体に係る経過的加算"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象人物, 対象期間, _parameters):
        世帯人数 = 対象人物.世帯("世帯人数", 対象期間)
        # インデックスは0始まりのため1を引いて調整
        世帯人数区分インデックス = np.clip(世帯人数, 1, 5) - 1

        居住級地区分1 = 対象人物.世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象人物.世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        年齢 = 対象人物("年齢", 対象期間)
        年齢区分インデックス = np.select(
            [(年齢 >= 0) * (年齢 <= 2),
             (年齢 >= 3) * (年齢 <= 5),
             (年齢 >= 6) * (年齢 <= 11),
             (年齢 >= 12) * (年齢 <= 17),
             (年齢 >= 18) * (年齢 <= 19),
             (年齢 >= 20) * (年齢 <= 40),
             (年齢 >= 41) * (年齢 <= 59),
             (年齢 >= 60) * (年齢 <= 64),
             (年齢 >= 65) * (年齢 <= 69),
             (年齢 >= 70) * (年齢 <= 74),
             (年齢 >= 75)],
            list(range(11)),
            -1)

        # NOTE: この行はentityを人物にして計算する必要がある
        # (世帯の場合、年齢区分の要素数が人物、居住級地区分の要素数が世帯となり配列の形式が一致せずエラーになる)
        加算額 = 生活扶助本体に係る経過的加算表()[世帯人数区分インデックス, 年齢区分インデックス, 居住級地区分インデックス]
        return 加算額


class 障害者加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "障害者に関する加算"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        # NOTE: 世帯員ごとの所属世帯の居住級地区分1を取得。世帯員ごとの身体障害者手帳等級と対応づけるため
        # (世帯ごとの場合、居住級地区分の要素数が世帯数、身体障害者手帳等級区分の要素数が人物数となり、配列の形式が一致せずエラーになる)
        居住級地区分1 = 対象世帯.members("居住級地区分1_世帯員", 対象期間)
        居住級地区分インデックス = np.select(
            [居住級地区分1 == 1,
             居住級地区分1 == 2,
             居住級地区分1 == 3],
            list(range(3)),
            -1).astype(int)

        身体障害者手帳等級 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        # 該当者(1級~3級)のみ抽出
        身体障害者手帳等級インデックス = np.select(
            [身体障害者手帳等級 == 身体障害者手帳等級パターン.一級,
             身体障害者手帳等級 == 身体障害者手帳等級パターン.二級,
             身体障害者手帳等級 == 身体障害者手帳等級パターン.三級],
            list(range(3)),
            -1)

        対象である = 身体障害者手帳等級インデックス != -1
        加算額 = 障害者加算表()[身体障害者手帳等級インデックス, 居住級地区分インデックス]

        return 対象世帯.sum(対象である * 加算額)


class 母子加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "母子世帯等に関する加算（父子世帯も対象）"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        配偶者がいない = 対象世帯.nb_persons(世帯.親) == 1

        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)

        児童である = 対象世帯.members("児童", 対象期間)
        児童の人数 = 対象世帯.sum(児童である)

        居住級地区分1の金額 = 配偶者がいない * np.select(
            [児童の人数 == 1, 児童の人数 == 2, 児童の人数 >= 3],
            [18800, 23600, 23600 + 2900 * (児童の人数 - 2)],
            0)

        居住級地区分2の金額 = 配偶者がいない * np.select(
            [児童の人数 == 1, 児童の人数 == 2, 児童の人数 >= 3],
            [17400, 21800, 21800 + 2700 * (児童の人数 - 2)],
            0)

        居住級地区分3の金額 = 配偶者がいない * np.select(
            [児童の人数 == 1, 児童の人数 == 2, 児童の人数 >= 3],
            [16100, 20200, 20200 + 2500 * (児童の人数 - 2)],
            0)

        return np.select(
            [居住級地区分1 == 1,
             居住級地区分1 == 2,
             居住級地区分1 == 3],
            [居住級地区分1の金額,
             居住級地区分2の金額,
             居住級地区分3の金額],
            -1).astype(int)


class 児童(Variable):
    value_type = bool
    entity = 人物
    default_value = False
    definition_period = DAY
    label = "児童かどうか"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象人物, 対象期間, parameters):
        # 児童とは、18歳になる日以後の最初の3月31日までの者
        年齢 = 対象人物("年齢", 対象期間)
        return np.select([年齢 < 18, 年齢 == 18, 年齢 > 18], [True, 対象期間.start.month <= 3, False], False)


class 児童を養育する場合の加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "児童を養育する場合の加算"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        児童である = 対象世帯.members("児童", 対象期間)
        児童の人数 = 対象世帯.sum(児童である)
        return 児童の人数 * 10190


class 母子世帯等に係る経過的加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "母子世帯等に係る経過的加算（父子世帯も対象）"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        年齢 = 対象世帯.members("年齢", 対象期間)
        子供の年齢降順 = 対象世帯.get_rank(対象世帯, -年齢, condition=対象世帯.has_role(世帯.子))
        # 子供の人数が1人であることは確認済みのため、第一子の年齢=子供の年齢となる
        第一子である = 子供の年齢降順 == 0
        子供の年齢 = 対象世帯.sum(第一子である * 年齢)

        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        世帯人数区分インデックス = np.select(
            [世帯人数 == 3,
             世帯人数 == 4,
             世帯人数 >= 5],
            list(range(3)),
            -1).astype(int)

        年齢区分インデックス = np.select(
            [(子供の年齢 >= 0) * (子供の年齢 <= 2),
             (子供の年齢 >= 3) * (子供の年齢 <= 5),
             (子供の年齢 >= 6) * (子供の年齢 <= 11),
             (子供の年齢 >= 12) * (子供の年齢 <= 14),
             (子供の年齢 >= 15) * (子供の年齢 <= 17),
             (子供の年齢 >= 18) * (子供の年齢 < 20)],
            list(range(6)),
            -1).astype(int)

        子供の人数 = 対象世帯.nb_persons(世帯.子)
        配偶者がいない = 対象世帯.nb_persons(世帯.親) == 1
        対象である = (子供の人数 == 1) * 配偶者がいない * (世帯人数区分インデックス != -1) * (年齢区分インデックス != -1)

        return 対象である * 母子世帯等に係る経過的加算表()[世帯人数区分インデックス, 年齢区分インデックス, 居住級地区分インデックス]


class 入院中(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "入院しているか否か"


class 児童を養育する場合に係る経過的加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "児童を養育する場合に係る経過的加算"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        # プライベートメソッドが使用できないため、別クラスに切り出し
        三人以下の世帯であって三歳未満の児童が入院している場合の経過的加算 = 対象世帯("三人以下の世帯であって三歳未満の児童が入院している場合の経過的加算", 対象期間)
        四人以上の世帯であって三歳未満の児童がいる場合の経過的加算 = 対象世帯("四人以上の世帯であって三歳未満の児童がいる場合の経過的加算", 対象期間)
        第三子以降の三歳から小学生修了前の児童がいる場合の経過的加算 = 対象世帯("第三子以降の三歳から小学生修了前の児童がいる場合の経過的加算", 対象期間)
        return (
            三人以下の世帯であって三歳未満の児童が入院している場合の経過的加算
            + 四人以上の世帯であって三歳未満の児童がいる場合の経過的加算
            + 第三子以降の三歳から小学生修了前の児童がいる場合の経過的加算
        )


class 三人以下の世帯であって三歳未満の児童が入院している場合の経過的加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "児童を養育する場合に係る経過的加算（３人以下の世帯であって、３歳未満の児童が入院している等の場合）"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        世帯人数 = 対象世帯("世帯人数", 対象期間)
        対象である = 世帯人数 <= 3

        各世帯員の年齢 = 対象世帯.members("年齢", 対象期間)
        各世帯員が3歳未満 = 各世帯員の年齢 < 3
        各世帯員が入院中 = 対象世帯.members("入院中", 対象期間)
        加算対象者数 = 対象世帯.sum(各世帯員が3歳未満 & 各世帯員が入院中)
        return 対象である * (加算対象者数 * 4330)


class 四人以上の世帯であって三歳未満の児童がいる場合の経過的加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "児童を養育する場合に係る経過的加算（４人以上の世帯であって、３歳未満の児童がいる場合）"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        世帯人数 = 対象世帯("世帯人数", 対象期間)
        対象である = 世帯人数 >= 4

        各世帯員の年齢 = 対象世帯.members("年齢", 対象期間)
        各世帯員が3歳未満かどうか = 各世帯員の年齢 < 3
        加算対象者数 = 対象世帯.sum(各世帯員が3歳未満かどうか)
        return 対象である * (加算対象者数 * 4330)


class 第三子以降の三歳から小学生修了前の児童がいる場合の経過的加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "児童を養育する場合に係る経過的加算（第３子以降の「３歳から小学生修了前」の児童がいる場合）"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        子供である = 対象世帯.has_role(世帯.子)

        年齢 = 対象世帯.members("年齢", 対象期間)
        子供の年齢の降順インデックス = 対象世帯.get_rank(対象世帯, -年齢, condition=子供である)

        # NOTE: インデックスは0始まりのため、0は第一子, 1は第二子を意味する。また、子でない場合-1となるため該当しない
        第三子以降である = 子供の年齢の降順インデックス > 1

        三歳以上である = 年齢 >= 3

        学年 = 対象世帯.members("学年", 対象期間)
        小学生終了前である = 学年 <= 6

        加算対象者数 = 対象世帯.sum(第三子以降である & 三歳以上である & 小学生終了前である)
        return 加算対象者数 * 4330


class 放射線障害者パターン(Enum):
    __order__ = "無 現罹患者 元罹患者"
    無 = "無"
    現罹患者 = "現罹患者"
    元罹患者 = "元罹患者"


class 放射線障害(Variable):
    value_type = Enum
    possible_values = 放射線障害者パターン
    default_value = 放射線障害者パターン.無
    entity = 人物
    definition_period = DAY
    label = "放射線障害状況"


class 放射線障害者加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "放射線障害者加算"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://seikatsuhogo.biz/blogs/105
    """

    def formula(対象世帯, 対象期間, parameters):
        各世帯員の放射線障害 = 対象世帯.members("放射線障害", 対象期間)
        現罹患者の人数 = 対象世帯.sum(各世帯員の放射線障害 == 放射線障害者パターン.現罹患者)
        元罹患者の人数 = 対象世帯.sum(各世帯員の放射線障害 == 放射線障害者パターン.元罹患者)

        return 現罹患者の人数 * 43830 + 元罹患者の人数 * 21920


class 妊産婦パターン(Enum):
    __order__ = "無 妊娠6ヵ月未満 妊娠6ヵ月以上 産後6ヵ月以内"
    無 = "無"
    妊娠6ヵ月未満 = "妊娠6ヵ月未満"
    妊娠6ヵ月以上 = "妊娠6ヵ月以上"
    産後6ヵ月以内 = "産後6ヵ月以内"


class 妊産婦(Variable):
    value_type = Enum
    possible_values = 妊産婦パターン
    default_value = 妊産婦パターン.無
    entity = 人物
    definition_period = DAY
    label = "妊産婦の妊娠、産後状況"


class 妊産婦加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "妊産婦加算"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://seikatsuhogo.biz/blogs/105
    """

    def formula(対象世帯, 対象期間, parameters):
        妊産婦 = 対象世帯.members("妊産婦", 対象期間)
        妊娠6ヵ月未満の人数 = 対象世帯.sum(妊産婦 == 妊産婦パターン.妊娠6ヵ月未満)
        妊娠6ヵ月以上の人数 = 対象世帯.sum(妊産婦 == 妊産婦パターン.妊娠6ヵ月以上)
        産後6ヵ月以内の人数 = 対象世帯.sum(妊産婦 == 妊産婦パターン.産後6ヵ月以内)

        return 妊娠6ヵ月未満の人数 * 9130 + 妊娠6ヵ月以上の人数 * 13790 + 産後6ヵ月以内の人数 * 8480


class 介護施設入所中(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "介護施設に入所しているか否か"


class 介護施設入所者加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "介護施設入所者加算"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://seikatsuhogo.biz/blogs/105
    """

    def formula(対象世帯, 対象期間, parameters):
        介護施設入所中 = 対象世帯.members("介護施設入所中", 対象期間)
        # NOTE: 世帯ごとに人数を数える必要があるため対象世帯.sumは使えない（Trueは1、Falseは0なのでsumを取ることで人数が分かる）
        加算対象者数 = 対象世帯.sum(介護施設入所中)
        return 加算対象者数 * 9880


class 在宅療養中(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "在宅で療養に専念している患者(結核又は3ヶ月以上の治療を要するもの)か否か"


class 在宅患者加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "在宅患者加算"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://seikatsuhogo.biz/blogs/105
    """

    def formula(対象世帯, 対象期間, parameters):
        各世帯員が在宅療養中 = 対象世帯.members("在宅療養中", 対象期間)
        加算対象者数 = 対象世帯.sum(各世帯員が在宅療養中)
        return 加算対象者数 * 13270


class 住宅扶助基準(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "住宅扶助基準"
    reference = "http://kobekoubora.life.coocan.jp/2021juutakufujo.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        # 算出できるのは上限額だが、額が大きいため加算する
        # NOTE: 母子家庭や障害、病気などで特定の病院の近くに住む必要があるといった場合には、特別加算分が計上される場合もあるが、
        # 判定条件不明のため一旦無視
        居住都道府県 = 対象世帯("居住都道府県", 対象期間)
        居住市区町村 = 対象世帯("居住市区町村", 対象期間)
        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        世帯人員区分 = np.select(
            [世帯人数 == 1,
             世帯人数 == 2,
             (世帯人数 >= 3) * (世帯人数 <= 5),
             世帯人数 == 6,
             世帯人数 >= 7],
            list(range(5)),
            -1).astype(int)

        市区町村名一覧 = 市ごとの住宅扶助限度額_キー一覧()
        居住市区町村区分 = np.select(
            [居住市区町村 == 市区町村 for 市区町村 in 市区町村名一覧],
            list(range(len(市区町村名一覧))),
            -1).astype(int)

        # p.4~6までの市に居住している場合はp.4~6を適用
        市ごとの住宅扶助限度金額 = 市ごとの住宅扶助限度額()[居住市区町村区分, 世帯人員区分]

        都道府県キー一覧 = 都道府県ごとの住宅扶助限度額_キー一覧()
        都道府県区分 = np.select(
            [(居住都道府県 == キー["都道府県"]) * (居住級地区分1 == キー["級地"]) for キー in 都道府県キー一覧],
            list(range(len(都道府県キー一覧))),
            -1).astype(int)

        # それ以外の市区町村に居住している場合はp.1~3を適用
        都道府県ごとの住宅扶助限度金額 = 都道府県ごとの住宅扶助限度額()[都道府県区分, 世帯人員区分]

        市ごとの住宅扶助限度額が適用される = 居住市区町村区分 != -1

        return (市ごとの住宅扶助限度額が適用される * 市ごとの住宅扶助限度金額) + (np.logical_not(市ごとの住宅扶助限度額が適用される) * 都道府県ごとの住宅扶助限度金額)


class 教育扶助基準(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "教育扶助基準"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        学年 = 対象世帯.members("学年", 対象期間)
        支給額一覧 = np.select(
            [(学年 >= 1) * (学年 <= 6), (学年 >= 7) * (学年 <= 9)],
            [2600, 5100],
            0)
        return 対象世帯.sum(支給額一覧)


class 高等学校等就学費(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "高等学校等就学費"
    reference = "https://www.mhlw.go.jp/content/000776372.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        学年 = 対象世帯.members("学年", 対象期間)
        就学費 = np.select(
            [(学年 >= 10) * (学年 <= 12)],
            [5300],
            0)
        return 対象世帯.sum(就学費)


class 冬季加算(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "冬季加算"
    reference = "https://www.mhlw.go.jp/file/05-Shingikai-12601000-Seisakutoukatsukan-Sanjikanshitsu_Shakaihoshoutantou/26102103_6.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        冬季である = 対象世帯("冬季", 対象期間)

        冬季加算地域区分1 = 対象世帯("冬季加算地域区分1", 対象期間)
        # インデックスは0始まり、区分は1始まりなので調整
        冬季加算地域区分インデックス = 冬季加算地域区分1 - 1

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        世帯人数区分インデックス = np.select(
            [世帯人数 < 10],
            # インデックスは0始まりなので調整
            [世帯人数 - 1],
            -1).astype(int)

        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        # 9人の場合の加算に人数ごとの加算を追加
        世帯人数十人以上の場合の加算金額 = 冬季加算表()[冬季加算地域区分インデックス, -2, 居住級地区分インデックス] +\
            (世帯人数 - 9) * 冬季加算表()[冬季加算地域区分インデックス, -1, 居住級地区分インデックス]

        加算金額 = np.select(
            [世帯人数 >= 10],
            [世帯人数十人以上の場合の加算金額],
            冬季加算表()[冬季加算地域区分インデックス, 世帯人数区分インデックス, 居住級地区分インデックス])

        return 冬季である * 加算金額


class 冬季加算地域区分1(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "冬季加算地域区分"
    reference = "https://www.mhlw.go.jp/file/05-Shingikai-12601000-Seisakutoukatsukan-Sanjikanshitsu_Shakaihoshoutantou/26102103_6.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, _parameters):
        居住都道府県 = 対象世帯("居住都道府県", 対象期間)
        地域区分表キー一覧 = 地域区分表_キー一覧()
        居住都道府県インデックス = np.select(
            [居住都道府県 == 都道府県 for 都道府県 in 地域区分表キー一覧],
            list(range(len(地域区分表キー一覧))),
            -1).astype(int)

        地域区分 = 地域区分表()[居住都道府県インデックス]

        # 票に含まれないものは区分6
        return np.select([居住都道府県インデックス != -1],
                         [地域区分],
                         6)


class 冬季(Variable):
    value_type = bool
    default_value = False
    entity = 世帯
    definition_period = DAY
    label = "冬季加算の対象となる期間内であるかどうか"
    reference = "https://www.mhlw.go.jp/file/05-Shingikai-12601000-Seisakutoukatsukan-Sanjikanshitsu_Shakaihoshoutantou/26102103_6.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://seikatsuhogo.biz/blogs/105
    """

    def formula(対象世帯, 対象期間, parameters):
        冬季加算地域区分1 = 対象世帯("冬季加算地域区分1", 対象期間)

        月 = 対象期間.date.month

        # 11~3月（I, II区は10~4月）を冬季とする
        return np.select([(冬季加算地域区分1 == 1) + (冬季加算地域区分1 == 2)],
                         [(月 <= 4) + (月 >= 10)],
                         (月 <= 3) + (月 >= 11))


class 期末一時扶助(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "期末一時扶助"
    # TODO: 計算式について、官公庁の参考リンクも追加
    reference = "https://seikatsuhogo.biz/blogs/61"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.mhlw.go.jp/content/12002000/000771098.pdf
    https://www.holos.jp/media/welfare-amount-of-money.php
    https://www.wam.go.jp/wamappl/bb16GS70.nsf/0/573310120867b50d4925743c00047cb4/$FILE/20080501_1shiryou5_1.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        # 対象期間が12月の時のみ支給
        対象月である = 対象期間.date.month == 12

        世帯人数 = 対象世帯("世帯人数", 対象期間)
        # インデックスは0始まりのため調整
        世帯人数区分インデックス = 世帯人数 - 1

        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)
        居住級地区分2 = 対象世帯("居住級地区分2", 対象期間)
        居住級地区分インデックス = np.select(
            [(居住級地区分1 == 1) * (居住級地区分2 == 1),
             (居住級地区分1 == 1) * (居住級地区分2 == 2),
             (居住級地区分1 == 2) * (居住級地区分2 == 1),
             (居住級地区分1 == 2) * (居住級地区分2 == 2),
             (居住級地区分1 == 3) * (居住級地区分2 == 1),
             (居住級地区分1 == 3) * (居住級地区分2 == 2)],
            list(range(6)),
            -1).astype(int)

        # 9人の場合の加算に人数ごとの加算を追加
        世帯人数十人以上の場合の金額 = 期末一時扶助表()[-2, 居住級地区分インデックス] +\
            (世帯人数 - 9) * 期末一時扶助表()[-1, 居住級地区分インデックス]

        # NOTE: out of boundにならないようclipでインデックスの上限値を設定（10人以上の場合もこの値自体は計算される）
        世帯人数九人以下の場合の金額 = 期末一時扶助表()[np.clip(世帯人数区分インデックス, 0, 9), 居住級地区分インデックス]

        金額 = np.select(
            [世帯人数 > 9],
            [世帯人数十人以上の場合の金額],
            世帯人数九人以下の場合の金額)
        return 対象月である * 金額


class 勤労控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "勤労控除"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.holos.jp/media/welfare-income-earn.php
    """

    def formula(対象世帯, 対象期間, parameters):
        基礎控除 = 対象世帯("生活保護基礎控除", 対象期間)
        新規就労控除 = 対象世帯("新規就労控除", 対象期間)
        未成年者控除 = 対象世帯("未成年者控除", 対象期間)
        return 基礎控除 + 新規就労控除 + 未成年者控除


class 生活保護基礎控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "生活保護における基礎控除"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    所得税、住民税における基礎控除とは異なる。
    算出方法は以下リンクも参考になる。
    https://www.holos.jp/media/welfare-income-earn.php
    https://www.city.chiba.jp/hokenfukushi/hogo/documents/r3shinkijunngakuhyou.pdf
    """

    def formula(対象世帯, 対象期間, parameters):
        # TODO: 収入として年金等も入力するようになったら、勤労収入のみ計算対象に入れるようにする
        収入 = 対象世帯.members("収入", 対象期間)
        # 収入を高い順にソート
        収入の降順インデックス = 対象世帯.get_rank(対象世帯, -収入)
        月収 = 収入 / 12

        # 世帯で最も収入が多い世帯員が1人目の対象者
        一人目の対象者である = 収入の降順インデックス == 0
        二人目以降の対象者である = 収入の降順インデックス != 0

        一人目の控除 = np.clip((月収 - 15000) * 0.1 + 15000, 0, 月収)
        # 2人目以降の計算式は https://www.city.chiba.jp/hokenfukushi/hogo/documents/r3shinkijunngakuhyou.pdf の基礎控除表を参考に作成
        二人目以降の控除 = np.clip(np.clip((月収 - 41000) * 0.085, 0, None) + 14960 + np.clip(月収 - 14960, 0, 40), 0, 月収)

        return 対象世帯.sum(一人目の対象者である * 一人目の控除 + 二人目以降の対象者である * 二人目以降の控除)


class 六か月以内に新規就労(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "6か月以内に新たに継続性のある職業に従事したかどうか"


class 新規就労控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "新規就労控除"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.holos.jp/media/welfare-income-earn.php
    https://www.mhlw.go.jp/stf/shingi/2r9852000001ifbg-att/2r9852000001ifii.pdf (額は現在と異なる部分あり)
    """

    def formula(対象世帯, 対象期間, parameters):
        六か月以内に新規就労 = 対象世帯.members("六か月以内に新規就労", 対象期間)
        対象者数 = 対象世帯.sum(六か月以内に新規就労)
        return 対象者数 * 11700


class 未成年者控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "未成年者控除"
    reference = "https://www.mhlw.go.jp/content/12002000/000771098.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.holos.jp/media/welfare-income-earn.php
    https://www.mhlw.go.jp/stf/shingi/2r9852000001ifbg-att/2r9852000001ifii.pdf (額は現在と異なる部分あり)
    """

    def formula(対象世帯, 対象期間, parameters):
        未成年 = 対象世帯.members("年齢", 対象期間) < parameters(対象期間).全般.成人年齢
        # TODO: 収入として年金等も入力するようになったら、勤労収入のみ計算対象に入れるようにする
        就労中 = 対象世帯.members("収入", 対象期間) > 0
        対象者数 = 対象世帯.sum(未成年 & 就労中)
        return 対象者数 * 11600



"""
血液凝固因子異常症の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class 先天性の血液凝固因子異常症である可能性がある(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "先天性の血液凝固因子異常症である可能性がある"
    documentation = """
    「先天性の血液凝固因子異常症である」の場合と異なり、「その他」の場合も含める
    """

    def formula(対象人物, 対象期間, _parameters):
        先天性の血液凝固因子異常症である = 対象人物("先天性の血液凝固因子異常症である", 対象期間)
        その他 = 対象人物("血液凝固因子異常症_その他", 対象期間)
        return 先天性の血液凝固因子異常症である + その他


class 先天性の血液凝固因子異常症である(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "先天性の血液凝固因子異常症である"

    def formula(対象人物, 対象期間, _parameters):
        第I因子欠乏症 = 対象人物("血液凝固因子異常症_第I因子欠乏症", 対象期間)
        第II因子欠乏症 = 対象人物("血液凝固因子異常症_第II因子欠乏症", 対象期間)
        第V因子欠乏症 = 対象人物("血液凝固因子異常症_第V因子欠乏症", 対象期間)
        第VII因子欠乏症 = 対象人物("血液凝固因子異常症_第VII因子欠乏症", 対象期間)
        第VIII因子欠乏症 = 対象人物("血液凝固因子異常症_第VIII因子欠乏症", 対象期間)
        第IX因子欠乏症 = 対象人物("血液凝固因子異常症_第IX因子欠乏症", 対象期間)
        第X因子欠乏症 = 対象人物("血液凝固因子異常症_第X因子欠乏症", 対象期間)
        第XI因子欠乏症 = 対象人物("血液凝固因子異常症_第XI因子欠乏症", 対象期間)
        第XII因子欠乏症 = 対象人物("血液凝固因子異常症_第XII因子欠乏症", 対象期間)
        第XIII因子欠乏症 = 対象人物("血液凝固因子異常症_第XIII因子欠乏症", 対象期間)
        フォンヴィルブランド病 = 対象人物("血液凝固因子異常症_フォンヴィルブランド病", 対象期間)

        return 第I因子欠乏症 + 第II因子欠乏症 + 第V因子欠乏症 + 第VII因子欠乏症 + 第VIII因子欠乏症 +\
            第IX因子欠乏症 + 第X因子欠乏症 + 第XI因子欠乏症 + 第XII因子欠乏症 + 第XIII因子欠乏症 + フォンヴィルブランド病


# NOTE: OpenFiscaでは現状世帯員がEnumの配列を持つことができないため、因子ごとにbool値で表現している
# TODO: 世帯員がEnumを複数保持できるようになったらEnumに書き直す


class 血液凝固因子異常症_第I因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第I因子欠乏症）"
    documentation = """
    フィブリノゲン欠乏症
    """


class 血液凝固因子異常症_第II因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第II因子欠乏症）"
    documentation = """
    プロトロビン欠乏症
    """


class 血液凝固因子異常症_第V因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第V因子欠乏症）"
    documentation = """
    不安定因子欠乏症
    """


class 血液凝固因子異常症_第VII因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（VII因子欠乏症）"
    documentation = """
    安定因子欠乏症
    """


class 血液凝固因子異常症_第VIII因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第VIII因子欠乏症）"
    documentation = """
    血友病A
    """


class 血液凝固因子異常症_第IX因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第IX因子欠乏症）"
    documentation = """
    血友病B
    """


class 血液凝固因子異常症_第X因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第X因子欠乏症）"
    documentation = """
    スチュアートプラウア欠乏症
    """


class 血液凝固因子異常症_第XI因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第XI因子欠乏症）"
    documentation = """
    PTA欠乏症
    """


class 血液凝固因子異常症_第XII因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第XII因子欠乏症）"
    documentation = """
    ヘイグマン因子欠乏症
    """


class 血液凝固因子異常症_第XIII因子欠乏症(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（第XIII因子欠乏症）"
    documentation = """
    フィブリン安定化因子欠乏症
    """


class 血液凝固因子異常症_フォンヴィルブランド病(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（フォン・ヴィルブランド病）"


class 血液凝固因子異常症_その他(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子異常症（その他）"
    documentation = """
    上記いずれの因子にも該当しない、どれに該当するか分からない、または因子異常症であるかどうか分からない場合に選ばれる
    """


class 血友病の可能性がある(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "血友病の可能性がある"
    documentation = """
    血友病である場合に加え、分からない場合も含まれる
    """

    def formula(対象人物, 対象期間, _parameters):
        第VIII因子欠乏症 = 対象人物("血液凝固因子異常症_第VIII因子欠乏症", 対象期間)
        第IX因子欠乏症 = 対象人物("血液凝固因子異常症_第IX因子欠乏症", 対象期間)
        その他 = 対象人物("血液凝固因子異常症_その他", 対象期間)
        return 第VIII因子欠乏症 + 第IX因子欠乏症 + その他  # 分からない場合も選ばれる


class 血友病である(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "血友病である"

    def formula(対象人物, 対象期間, _parameters):
        第VIII因子欠乏症 = 対象人物("血液凝固因子異常症_第VIII因子欠乏症", 対象期間)
        第IX因子欠乏症 = 対象人物("血液凝固因子異常症_第IX因子欠乏症", 対象期間)
        return 第VIII因子欠乏症 + 第IX因子欠乏症



"""
C型肝炎の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class C型肝炎ウイルスに感染している(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "C型肝炎ウイルスに感染している"


class 血液製剤の投与によってC型肝炎ウイルスに感染した(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液製剤の投与によってC型肝炎ウイルスに感染した"


# TODO: 必要であれば内容ごとに分割する
class 肝硬変や肝がんに罹患しているまたは肝移植をおこなった(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "肝硬変、肝がんに罹患している、または肝移植をおこなった"



"""
災害障害見舞金の実装
"""


from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物


class 災害救助法の適用地域である(Variable):
    value_type = bool
    entity = 世帯
    definition_period = DAY
    label = "災害救助法の適用地域であるかどうか"
    reference = "https://elaws.e-gov.go.jp/document?lawid=322AC0000000118"
    documentation = """
    自然災害ごとに、市区町村ごとに適用有無が決まる
    災害発生時は適用有無が変化していくので注意
    適用状況 https://www.bousai.go.jp/taisaku/kyuujo/kyuujo_tekiyou.html
    """
    # TODO: 居住市区町村から災害救助法が適用されているかどうかを計算できるようにする


class 災害障害見舞金_最大(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "災害障害見舞金の最大額"
    reference = "https://www.bousai.go.jp/taisaku/choui/pdf/siryo1-1.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.city.hino.lg.jp/kurashi/annzen/hisai/1011473.html
    """

    def formula(対象世帯, 対象期間, parameters):
        災害救助法の適用地域である = 対象世帯("災害救助法の適用地域である", 対象期間)

        災害による重い後遺障害がある = 対象世帯.members("災害による重い後遺障害がある", 対象期間)

        生計維持者への支給額 = parameters(対象期間).災害.支援.災害障害見舞金.生計維持者への支給額
        生計維持者以外への支給額 = parameters(対象期間).災害.支援.災害障害見舞金.生計維持者以外への支給額

        所得一覧 = 対象世帯.members("所得", 対象期間)
        所得降順 = 対象世帯.get_rank(対象世帯, -所得一覧)

        # NOTE: 厳密な生計維持者の判定には被災前の各世帯員の所得と比較が必要だが、簡便のため所得が最も多い世帯員を生計維持者の候補とする
        # （災害関連の計算で入力される「収入」は「被災前の収入」を指すため、被災による収入変化は影響しない）
        # また、世帯所得制限額についても柔軟な対応が取られているため計算式には入れない
        # 災害弔慰金等、災害援護資金関係 https://www.bousai.go.jp/taisaku/kyuujo/pdf/h30kaigi/siryo3-1.pdf
        生計維持者である = 所得降順 == 0
        生計維持者以外 = 所得降順 != 0
        対象生計維持者人数 = 対象世帯.sum(災害による重い後遺障害がある * 生計維持者である)
        対象生計維持者以外人数 = 対象世帯.sum(災害による重い後遺障害がある * 生計維持者以外)
        支給額 = 対象生計維持者人数 * 生計維持者への支給額 + 対象生計維持者以外人数 * 生計維持者以外への支給額

        return 災害救助法の適用地域である * 支給額


class 災害障害見舞金_最小(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "災害障害見舞金の最小額"
    reference = "https://www.bousai.go.jp/taisaku/choui/pdf/siryo1-1.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.city.hino.lg.jp/kurashi/annzen/hisai/1011473.html
    """

    def formula(対象世帯, 対象期間, parameters):
        災害救助法の適用地域である = 対象世帯("災害救助法の適用地域である", 対象期間)

        災害による重い後遺障害がある = 対象世帯.members("災害による重い後遺障害がある", 対象期間)
        生計維持者以外への支給額 = parameters(対象期間).災害.支援.災害障害見舞金.生計維持者以外への支給額

        支給額 = 対象世帯.sum(災害による重い後遺障害がある) * 生計維持者以外への支給額
        return 災害救助法の適用地域である * 支給額


class 災害による重い後遺障害がある(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "災害による重い後遺障害がある"
    reference = "https://www.city.hino.lg.jp/kurashi/annzen/hisai/1011473.html"



"""
重度心身障害者医療費助成制度の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.療育手帳 import 療育手帳等級パターン
from openfisca_japan.variables.障害.精神障害者保健福祉手帳 import 精神障害者保健福祉手帳等級パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 重度心身障害者医療費助成制度の対象者がいる(Variable):
    value_type = bool
    entity = 世帯
    definition_period = DAY
    label = "重度心身障害者医療費助成制度の対象者がいるか否か"
    reference = "https://pap-net.jp/medical_bills/subsidy/"

    def formula(対象世帯, 対象期間, _parameters):
        重度心身障害者医療費助成制度の対象者である = 対象世帯.members("重度心身障害者医療費助成制度の対象者である", 対象期間)
        return 対象世帯.sum(重度心身障害者医療費助成制度の対象者である)


class 重度心身障害者医療費助成制度の対象者である(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "重度心身障害者医療費助成制度の対象者である可能性があるか否か"
    reference = "https://pap-net.jp/medical_bills/subsidy/"
    documentation = """
    都道府県によって条件や助成額が異なるため、対象か否かのみ計算
    居住地に対応した条件確認はせず、いずれかの都道府県の条件に該当する場合はTrueとする

    （参考）
    東京都: https://www.fukushi.metro.tokyo.lg.jp/seikatsu/josei/marusyo.html
    千葉県: https://www.pref.chiba.lg.jp/shoufuku/service/iryou/juudo.html
    埼玉県: https://www.pref.saitama.lg.jp/a0702/jyuudo.html
    山梨県: https://www.pref.yamanashi.jp/shogai-fks/judo/jidoukanpu-ikou.html
    """

    def formula(対象人物, 対象期間, _parameters):
        年齢 = 対象人物("年齢", 対象期間)
        身体障害者手帳等級 = 対象人物("身体障害者手帳等級", 対象期間)
        精神障害者保健福祉手帳等級 = 対象人物("精神障害者保健福祉手帳等級", 対象期間)
        愛の手帳等級 = 対象人物("愛の手帳等級", 対象期間)
        療育手帳等級 = 対象人物("療育手帳等級", 対象期間)
        障害基礎年金_等級_最大 = 対象人物("障害基礎年金_等級_最大", 対象期間)

        身体障害条件 = (身体障害者手帳等級 == 身体障害者手帳等級パターン.一級) +\
            (身体障害者手帳等級 == 身体障害者手帳等級パターン.二級) +\
            (身体障害者手帳等級 == 身体障害者手帳等級パターン.三級)
        知的障害条件 = (愛の手帳等級 == 愛の手帳等級パターン.一度) +\
            (愛の手帳等級 == 愛の手帳等級パターン.二度) +\
            (療育手帳等級 == 療育手帳等級パターン.A) +\
            (療育手帳等級 == 療育手帳等級パターン.B)
        精神障害条件 = (精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.一級) +\
            (精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.二級)

        # https://www.pref.saitama.lg.jp/a0702/jyuudo.html
        # 後期高齢者医療制度の障害認定を受けている方
        # (上記の障害条件に含まれない条件を追加)
        # https://www.saitama-koukikourei.org/seido/taisho/hihokensya/
        後期高齢者医療制度の障害認定年齢条件 = (年齢 >= 65) * (年齢 <= 74)
        後期高齢者医療制度の障害認定基準 = (身体障害者手帳等級 == 身体障害者手帳等級パターン.四級) +\
            (障害基礎年金_等級_最大 == 1) +\
            (障害基礎年金_等級_最大 == 2)
        障害認定条件 = 後期高齢者医療制度の障害認定年齢条件 * 後期高齢者医療制度の障害認定基準

        return 身体障害条件 + 知的障害条件 + 精神障害条件 + 障害認定条件



"""
This file defines a reform.

A reform is a set of modifications to be applied to a reference tax and benefit system to carry out experiments.

See https://openfisca.org/doc/key-concepts/reforms.html
"""

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from openfisca_core.parameters import Bracket
from openfisca_core.reforms import Reform


class modify_social_security_taxation(Reform):
    def apply(self):
        """
        Apply reform.

        A reform always defines an `apply` method that builds the reformed tax and benefit system from the reference one.
        See https://openfisca.org/doc/coding-the-legislation/reforms.html#writing-a-reform

        Our reform modifies the `社会保険料` parameter, which is a scale.
        This parameter is declared in `parameters/税金/社会保険料.yaml`.

        See https://openfisca.org/doc/coding-the-legislation/legislation_parameters.html
        """
        self.modify_parameters(modifier_function = self.modify_brackets)

    @staticmethod
    def modify_brackets(parameters):
        """
        Social security taxation reform.

        This function takes an argument `parameters` which is a in-memory representation
        of the YAML parameters. It can be modified and must be returned.
        """
        # Access the right parameter node:
        brackets = parameters.税金.社会保険料.brackets

        # Add 0.1 to the rates of the second bracket, keeping the same thresholds:
        for rate in brackets[1].rate.values_list:
            rate.value += 0.1

        # Remove the first bracket:
        del brackets[0]

        # Add a new bracket with a higher tax rate for rich people:
        new_bracket = Bracket(
            "new_bracket",
            data = {
                "rate": {"2017-01-01-01": {"value": 0.4}},
                "threshold": {"2017-01-01-01": {"value": 40000}},
                },
            )
        brackets.append(new_bracket)

        return parameters



"""
先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業の実装
"""

from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯


class 先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業_最大(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業の調査研究協力謝金（最大額）"
    reference = "https://www.pmda.go.jp/relief-services/health-welfare/0006.html"
    documentation = """
    謝金の金額については https://www.med.niigata-u.ac.jp/ifc/tebiki/blood/05.html
    """

    def formula(対象世帯, 対象期間, parameters):
        先天性の血液凝固因子異常症である可能性がある = 対象世帯.members("先天性の血液凝固因子異常症である可能性がある", 対象期間)
        C型肝炎ウイルスに感染している = 対象世帯.members("C型肝炎ウイルスに感染している", 対象期間)
        血液製剤の投与によってC型肝炎ウイルスに感染した = 対象世帯.members("血液製剤の投与によってC型肝炎ウイルスに感染した", 対象期間)
        肝硬変や肝がんに罹患しているまたは肝移植をおこなった = 対象世帯.members("肝硬変や肝がんに罹患しているまたは肝移植をおこなった", 対象期間)
        対象者である = 先天性の血液凝固因子異常症である可能性がある \
            * C型肝炎ウイルスに感染している \
            * 血液製剤の投与によってC型肝炎ウイルスに感染した \
            * 肝硬変や肝がんに罹患しているまたは肝移植をおこなった

        謝金 = parameters(対象期間).福祉.感染症.先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業.調査研究協力謝金
        return 対象世帯.sum(対象者である) * 謝金


class 先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業_最小(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業の調査研究協力謝金（最小額）"
    reference = "https://www.pmda.go.jp/relief-services/health-welfare/0006.html"
    documentation = """
    謝金の金額については https://www.med.niigata-u.ac.jp/ifc/tebiki/blood/05.html
    """

    def formula(対象世帯, 対象期間, parameters):
        先天性の血液凝固因子異常症である = 対象世帯.members("先天性の血液凝固因子異常症である", 対象期間)
        C型肝炎ウイルスに感染している = 対象世帯.members("C型肝炎ウイルスに感染している", 対象期間)
        血液製剤の投与によってC型肝炎ウイルスに感染した = 対象世帯.members("血液製剤の投与によってC型肝炎ウイルスに感染した", 対象期間)
        肝硬変や肝がんに罹患しているまたは肝移植をおこなった = 対象世帯.members("肝硬変や肝がんに罹患しているまたは肝移植をおこなった", 対象期間)
        対象者である = 先天性の血液凝固因子異常症である \
            * C型肝炎ウイルスに感染している \
            * 血液製剤の投与によってC型肝炎ウイルスに感染した \
            * 肝硬変や肝がんに罹患しているまたは肝移植をおこなった

        謝金 = parameters(対象期間).福祉.感染症.先天性の傷病治療によるC型肝炎患者に係るQOL向上等のための調査研究事業.調査研究協力謝金
        return 対象世帯.sum(対象者である) * 謝金



"""
特別障害者手当の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.障害.愛の手帳 import 愛の手帳等級パターン
from openfisca_japan.variables.障害.療育手帳 import 療育手帳等級パターン
from openfisca_japan.variables.障害.精神障害者保健福祉手帳 import 精神障害者保健福祉手帳等級パターン
from openfisca_japan.variables.障害.身体障害者手帳 import 身体障害者手帳等級パターン


class 特別障害者手当_最大(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "特別障害者手当の最大額"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/tokubetsu.html"
    documentation = """
    厳密な判定には詳細な症状が必要なため、身体障害者手帳、精神障害者保健福祉手帳等から推定可能な最小値、最大値を算出

    算出方法は以下リンクも参考になる。
    https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html
    """

    def formula(対象世帯, 対象期間, parameters):
        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 >= 20

        身体障害者手帳等級 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        精神障害者保健福祉手帳等級 = 対象世帯.members("精神障害者保健福祉手帳等級", 対象期間)
        愛の手帳等級 = 対象世帯.members("愛の手帳等級", 対象期間)
        療育手帳等級 = 対象世帯.members("療育手帳等級", 対象期間)

        # 身体障害者手帳2級以上、療育手帳（愛の手帳）2度以上または精神障害者保健福祉手帳1級相当の障害が重複していることが条件のため
        # いずれかを持っている場合は最大値を満額とする
        # https://www.atgp.jp/knowhow/oyakudachi/c5943/
        身体障害条件 = (身体障害者手帳等級 == 身体障害者手帳等級パターン.一級) + (身体障害者手帳等級 == 身体障害者手帳等級パターン.二級)
        # NOTE: 療育手帳、愛の手帳いずれを指定した場合も該当するよう判定（療育手帳Aが愛の手帳1度~2度に相当）
        # https://www.rease.e.u-tokyo.ac.jp/read/jp/archive/statistics/leaf11031504p04.pdf
        知的障害条件 = ((愛の手帳等級 == 愛の手帳等級パターン.一度) + (愛の手帳等級 == 愛の手帳等級パターン.二度)
            + (療育手帳等級 == 療育手帳等級パターン.A))
        精神障害条件 = 精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.一級

        障害条件 = 身体障害条件 + (精神障害条件 + 知的障害条件)

        所得条件 = 対象世帯.members("特別障害者手当所得制限", 対象期間)

        # TODO: 介護施設以外の施設に入所中の場合も考慮する
        # 3か月以上継続して入院している場合のみ対象外となるため、最小額では「入院中」条件を参照せず満額とする
        入院入所条件 = np.logical_not(対象世帯.members("介護施設入所中", 対象期間))

        人物ごとの受給条件 = 年齢条件 * 障害条件 * 所得条件 * 入院入所条件
        対象人数 = 対象世帯.sum(人物ごとの受給条件)

        return 対象人数 * parameters(対象期間).福祉.特別障害者手当.支給額


class 特別障害者手当_最小(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "特別障害者手当の最小額"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/tokubetsu.html"
    documentation = """
    厳密な判定には詳細な症状が必要なため、身体障害者手帳、精神障害者保健福祉手帳等から推定可能な最小値、最大値を算出

    算出方法は以下リンクも参考になる。
    https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html
    https://h-navi.jp/column/article/35029230
    """

    def formula(対象世帯, 対象期間, parameters):
        年齢 = 対象世帯.members("年齢", 対象期間)
        年齢条件 = 年齢 >= 20

        身体障害者手帳等級 = 対象世帯.members("身体障害者手帳等級", 対象期間)
        精神障害者保健福祉手帳等級 = 対象世帯.members("精神障害者保健福祉手帳等級", 対象期間)
        愛の手帳等級 = 対象世帯.members("愛の手帳等級", 対象期間)
        療育手帳等級 = 対象世帯.members("療育手帳等級", 対象期間)

        # 身体障害者手帳2級以上、療育手帳（愛の手帳）2度以上または精神障害者保健福祉手帳1級相当の障害が重複していることが条件のため
        # いずれかを持っている場合は最大値を満額とする
        # https://www.atgp.jp/knowhow/oyakudachi/c5943/
        身体障害条件 = (身体障害者手帳等級 == 身体障害者手帳等級パターン.一級) + (身体障害者手帳等級 == 身体障害者手帳等級パターン.二級)
        # NOTE: 療育手帳、愛の手帳いずれを指定した場合も該当するよう判定（療育手帳Aが愛の手帳1度~2度に相当）
        # https://www.rease.e.u-tokyo.ac.jp/read/jp/archive/statistics/leaf11031504p04.pdf
        知的障害条件 = ((愛の手帳等級 == 愛の手帳等級パターン.一度) + (愛の手帳等級 == 愛の手帳等級パターン.二度)
            + (療育手帳等級 == 療育手帳等級パターン.A))
        精神障害条件 = 精神障害者保健福祉手帳等級 == 精神障害者保健福祉手帳等級パターン.一級

        障害条件 = 身体障害条件 * (精神障害条件 + 知的障害条件)

        所得条件 = 対象世帯.members("特別障害者手当所得制限", 対象期間)

        # TODO: 介護施設以外の施設に入所中の場合も考慮する
        入院入所条件 = np.logical_not(対象世帯.members("入院中", 対象期間) + 対象世帯.members("介護施設入所中", 対象期間))

        人物ごとの受給条件 = 年齢条件 * 障害条件 * 所得条件 * 入院入所条件
        対象人数 = 対象世帯.sum(人物ごとの受給条件)

        return 対象人数 * parameters(対象期間).福祉.特別障害者手当.支給額


class 特別障害者手当所得制限(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "特別障害者手当の所得制限"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/tokubetsu.html"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html
    """

    def formula(対象人物, 対象期間, parameters):
        特別障害者手当 = parameters(対象期間).福祉.特別障害者手当
        扶養人数 = 対象人物.世帯("扶養人数", 対象期間)
        # 複数世帯入力(2以上の長さのndarray入力)対応のためndarray化
        本人所得を参照する場合の所得制限限度額 = np.array(特別障害者手当.所得制限限度額.本人)[np.clip(扶養人数, 0, 9)]
        扶養義務者所得を参照する場合の所得制限限度額 = np.array(特別障害者手当.所得制限限度額.扶養義務者)[np.clip(扶養人数, 0, 9)]

        # 加算額を考慮
        # NOTE: += は使用不可（Variableの破壊的変更により正しく計算されない）
        本人所得を参照する場合の所得制限限度額 = \
            本人所得を参照する場合の所得制限限度額 + 対象人物.世帯("特別障害者手当_受給者の所得加算額", 対象期間)
        扶養義務者所得を参照する場合の所得制限限度額 = \
            扶養義務者所得を参照する場合の所得制限限度額 + (扶養人数 >= 2) * 対象人物.世帯("特別障害者手当_扶養義務者の所得加算額", 対象期間)

        本人の控除後所得 = 対象人物("特別障害者手当の控除後所得_本人所得", 対象期間)
        # 便宜上、世帯の最大の所得を扶養義務者の所得とする
        扶養義務者の控除後所得 = 対象人物.世帯.max(対象人物("特別障害者手当の控除後所得_扶養義務者所得", 対象期間))

        # NOTE: 本人所得の場合は限度額以下、扶養義務者所得の場合は限度額より少ない場合に支給
        本人の所得制限 = 本人の控除後所得 <= 本人所得を参照する場合の所得制限限度額
        扶養義務者の所得制限 = 扶養義務者の控除後所得 < 扶養義務者所得を参照する場合の所得制限限度額

        return 本人の所得制限 * 扶養義務者の所得制限


class 特別障害者手当の控除後所得_扶養義務者所得(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "各種控除が適用された後の特別障害者手当における世帯高所得額（扶養義務者所得を計算する場合）"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html"

    def formula(対象人物, 対象期間, parameters):
        # TODO: 障害年金実装時に所得に加算する
        # (障害年金、遺族年金等の公的年金も所得に加算される)
        所得 = 対象人物("所得", 対象期間)

        給与所得及び雑所得からの控除額 = parameters(対象期間).所得.給与所得及び雑所得からの控除額
        社会保険料控除 = parameters(対象期間).所得.社会保険料相当額

        # 上限33万円
        配偶者特別控除 = np.clip(対象人物("配偶者特別控除", 対象期間), 0, 330000)

        障害者控除 = 対象人物("障害者控除", 対象期間)
        寡婦控除 = 対象人物("寡婦控除", 対象期間)
        ひとり親控除 = 対象人物("ひとり親控除", 対象期間)
        勤労学生控除 = 対象人物("勤労学生控除", 対象期間)

        # 他の控除（雑損控除・医療費控除等）は定額でなく実費を元に算出するため除外する
        控除総額 = 給与所得及び雑所得からの控除額 + 社会保険料控除 + 配偶者特別控除 + 障害者控除 + 寡婦控除 + ひとり親控除 + 勤労学生控除
        控除後所得 = 所得 - 控除総額

        return 控除後所得


class 特別障害者手当の控除後所得_本人所得(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "各種控除が適用された後の特別障害者手当における世帯高所得額（本人所得を計算する場合）"
    reference = "https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html"

    def formula(対象人物, 対象期間, parameters):
        # TODO: 障害年金実装時に所得に加算する
        # (障害年金、遺族年金等の公的年金も所得に加算される)
        所得 = 対象人物("所得", 対象期間)

        給与所得及び雑所得からの控除額 = parameters(対象期間).所得.給与所得及び雑所得からの控除額

        # 本人所得に対しては社会保険料控除0円で計算される
        社会保険料控除 = 0
        # 上限33万円
        配偶者特別控除 = np.clip(対象人物("配偶者特別控除", 対象期間), 0, 330000)

        障害者控除 = 対象人物("障害者控除", 対象期間)

        # 本人所得の場合、障害者控除, 特別障害者控除については0円で計算
        障害者控除対象 = 対象人物("障害者控除対象", 対象期間)
        特別障害者控除対象 = 対象人物("特別障害者控除対象", 対象期間)
        本人分の障害者控除 = np.select(
            [障害者控除対象, 特別障害者控除対象],
            [parameters(対象期間).所得.障害者控除額, parameters(対象期間).所得.特別障害者控除額],
            0)
        # 本人分の障害者控除、特別障害者控除は0円で計算される
        障害者控除 = np.clip(障害者控除 - 本人分の障害者控除, 0, None)

        # TODO: 同居特別障害者の場合も通常の特別障害者と同じ額に補正する必要がある？
        # 単純に記載を省略しているだけの可能性もあり (https://www.fukushi.metro.tokyo.lg.jp/shinsho/teate/toku_shou.html で国の制度を照会しているが同居特別障害者の記載はない)

        寡婦控除 = 対象人物("寡婦控除", 対象期間)
        ひとり親控除 = 対象人物("ひとり親控除", 対象期間)
        勤労学生控除 = 対象人物("勤労学生控除", 対象期間)

        # 他の控除（雑損控除・医療費控除等）は定額でなく実費を元に算出するため除外する
        控除総額 = 給与所得及び雑所得からの控除額 + 社会保険料控除 + 配偶者特別控除 + 障害者控除 + 寡婦控除 + ひとり親控除 + 勤労学生控除
        控除後所得 = 所得 - 控除総額

        return 控除後所得


class 特別障害者手当_受給者の所得加算額(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "特別障害者手当_受給者の所得加算額"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/tokubetsu.html"

    def formula(対象世帯, 対象期間, parameters):
        年齢 = 対象世帯.members("年齢", 対象期間)
        # 同一生計配偶者のうち70歳以上の者
        同一生計配偶者である = 対象世帯.members("同一生計配偶者である", 対象期間)
        配偶者が70歳以上 = 年齢 >= 70
        配偶者70歳以上の人数 = 対象世帯.sum(同一生計配偶者である * 配偶者が70歳以上)

        扶養親族である = 対象世帯.members("扶養親族である", 対象期間)

        # 老人扶養親族
        老人扶養親族である = 扶養親族である * (年齢 >= 70)
        老人扶養親族の人数 = 対象世帯.sum(老人扶養親族である)

        # 特定扶養親族又は控除対象扶養親族（19歳未満の者に限る。）
        # 特定扶養親族と控除対象扶養親族の定義: https://www.nta.go.jp/taxes/shiraberu/taxanswer/yogo/senmon.htm#word9
        # TODO: 特定扶養親族と控除対象扶養親族をvariableに切り出す
        特定扶養親族である = 扶養親族である * (年齢 >= 19) * (年齢 < 23)
        控除対象扶養親族である = 扶養親族である * (年齢 >= 16)
        特定扶養親族又は控除対象扶養親族である = 特定扶養親族である + (控除対象扶養親族である * (年齢 < 19))
        特定扶養親族又は控除対象扶養親族の人数 = 対象世帯.sum(特定扶養親族又は控除対象扶養親族である)

        所得制限限度額 = parameters(対象期間).福祉.特別障害者手当.所得制限限度額
        老人扶養親族加算額 = (配偶者70歳以上の人数 + 老人扶養親族の人数) * 所得制限限度額.本人_老人扶養親族加算額
        特定扶養親族加算額 = 特定扶養親族又は控除対象扶養親族の人数 * 所得制限限度額.本人_特定扶養親族加算額

        return 老人扶養親族加算額 + 特定扶養親族加算額


class 特別障害者手当_扶養義務者の所得加算額(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "特別障害者手当_扶養義務者の所得加算額"
    reference = "https://www.mhlw.go.jp/bunya/shougaihoken/jidou/tokubetsu.html"

    def formula(対象世帯, 対象期間, parameters):
        年齢 = 対象世帯.members("年齢", 対象期間)

        # 老人扶養親族
        老人扶養親族である = 対象世帯.members("扶養親族である", 対象期間) * (年齢 >= 70)
        老人扶養親族の人数 = 対象世帯.sum(老人扶養親族である)

        # 老人扶養親族以外の人数
        老人扶養親族以外の扶養親族の人数 = 対象世帯.sum(対象世帯.members("扶養親族である", 対象期間) * (np.logical_not(老人扶養親族である)))

        # 老人扶養親族のみの場合、1人を除いた人数を加算
        # それ以外の場合、老人扶養親族の人数を加算
        加算人数 = np.where(老人扶養親族以外の扶養親族の人数 == 0, 老人扶養親族の人数 - 1, 老人扶養親族の人数)

        return 加算人数 * parameters(対象期間).福祉.特別障害者手当.所得制限限度額.扶養義務者_老人扶養親族加算額



"""
災害弔慰金の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯


class 災害弔慰金(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "災害弔慰金"
    reference = "https://www.bousai.go.jp/taisaku/choui/pdf/siryo1-1.pdf"
    documentation = """
    算出方法は以下リンクも参考になる。
    https://www.city.hino.lg.jp/kurashi/annzen/hisai/1011473.html
    """

    def formula(対象世帯, 対象期間, parameters):
        災害救助法の適用地域である = 対象世帯("災害救助法の適用地域である", 対象期間)

        災害で死亡した世帯員の人数 = 対象世帯("災害で死亡した世帯員の人数", 対象期間)
        災害で生計維持者が死亡した = 対象世帯("災害で生計維持者が死亡した", 対象期間)

        生計維持者死亡の場合の支給額 = parameters(対象期間).災害.支援.災害弔慰金.生計維持者死亡の場合の支給額
        生計維持者以外死亡の場合の支給額 = parameters(対象期間).災害.支援.災害弔慰金.生計維持者以外死亡の場合の支給額

        支給額 = np.select(
            [災害で生計維持者が死亡した],
            [生計維持者死亡の場合の支給額 + (災害で死亡した世帯員の人数 - 1) * 生計維持者以外死亡の場合の支給額],
            災害で死亡した世帯員の人数 * 生計維持者以外死亡の場合の支給額).astype(int)

        return 災害救助法の適用地域である * 支給額


class 災害で死亡した世帯員の人数(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "災害で死亡した世帯員の人数"
    reference = "https://www.bousai.go.jp/taisaku/choui/pdf/siryo1-1.pdf"
    documentation = """
    死亡した世帯員は世帯情報に含められない（含めると他の制度の対象になってしまう）ため、死亡者に関するVariableを別途用意
    """


class 災害で生計維持者が死亡した(Variable):
    value_type = bool
    entity = 世帯
    definition_period = DAY
    label = "災害で生計維持者が死亡したかどうか"
    reference = "https://www.bousai.go.jp/taisaku/choui/pdf/siryo1-1.pdf"
    documentation = """
    死亡した世帯員は世帯情報に含められない（含めると他の制度の対象になってしまう）ため、死亡者に関するVariableを別途用意
    """



"""
健康管理費用の実装
"""

import numpy as np
from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 世帯, 人物


class 健康管理費用_最大(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "健康管理費用の最大額"
    reference = "https://www.pmda.go.jp/relief-services/hiv-positives/0002.html"
    documentation = """
    CD4(T4)リンパ球の量によって支給額が異なるが、variableで表す代わりに幅を持たせて算出
    """

    def formula(対象世帯, 対象期間, parameters):
        対象者である = 対象世帯.members("健康管理費用の対象である", 対象期間)
        支給額 = parameters(対象期間).福祉.感染症.健康管理費用.支給額_最大
        return 対象世帯.sum(対象者である) * 支給額


class 健康管理費用_最小(Variable):
    value_type = int
    entity = 世帯
    definition_period = DAY
    label = "健康管理費用の最小額"
    reference = "https://www.pmda.go.jp/relief-services/hiv-positives/0002.html"
    documentation = """
    CD4(T4)リンパ球の量によって支給額が異なるが、variableで表す代わりに幅を持たせて算出
    """

    def formula(対象世帯, 対象期間, parameters):
        対象者である = 対象世帯.members("健康管理費用の対象である", 対象期間)
        支給額 = parameters(対象期間).福祉.感染症.健康管理費用.支給額_最小
        return 対象世帯.sum(対象者である) * 支給額


class 健康管理費用の対象である(Variable):
    value_type = bool
    entity = 人物
    definition_period = DAY
    label = "健康管理費用の対象であるか否か"
    reference = "https://www.pmda.go.jp/relief-services/hiv-positives/0002.html"
    documentation = """
    CD4(T4)リンパ球の量によって支給額が異なるが、variableで表す代わりに幅を持たせて算出
    """

    def formula(対象人物, 対象期間, _parameters):
        HIV感染者である = 対象人物("HIV感染者である", 対象期間)
        血液製剤の投与によってHIVに感染した = 対象人物("血液製剤の投与によってHIVに感染した", 対象期間)
        家族に血液製剤によるHIV感染者がいる = 対象人物("家族に血液製剤によるHIV感染者がいる", 対象期間)
        エイズを発症している = 対象人物("エイズを発症している", 対象期間)

        対象者である = HIV感染者である * (血液製剤の投与によってHIVに感染した + 家族に血液製剤によるHIV感染者がいる) \
            * np.logical_not(エイズを発症している)
        return 対象者である



"""
HIV関連情報の実装
"""


from openfisca_core.periods import DAY
from openfisca_core.variables import Variable
from openfisca_japan.entities import 人物


class HIV感染者である(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "HIV感染者である"


class 血液製剤の投与によってHIVに感染した(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "血液凝固因子製剤の投与によってHIVに感染した"


class 家族に血液製剤によるHIV感染者がいる(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "家族に血液凝固因子製剤によるHIV感染者がいる"


class エイズを発症している(Variable):
    value_type = bool
    default_value = False
    entity = 人物
    definition_period = DAY
    label = "エイズを発症している"



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a 人物, a 世帯…

See https://openfisca.org/doc/key-concepts/variables.html
"""

from functools import cache

import numpy as np
from openfisca_core.periods import DAY, period
from openfisca_core.variables import Variable
from openfisca_japan import COUNTRY_DIR
# Import the Entities specifically defined for this tax and benefit system
from openfisca_japan.entities import 世帯, 人物
from openfisca_japan.variables.全般 import 性別パターン

# NOTE: 項目数が多い金額表は可読性の高いCSV形式としている。


@cache
def 配偶者控除額表():
    """
    csvファイルから値を取得

    配偶者控除額表()[配偶者の所得区分, 納税者本人の所得区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/住民税/配偶者控除額.csv",
                  delimiter=",", skip_header=1, dtype="int64")[np.newaxis, 1:]


@cache
def 老人控除対象配偶者_配偶者控除額表():
    """
    csvファイルから値を取得

    老人控除対象配偶者_配偶者控除額表()[配偶者の所得区分, 納税者本人の所得区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/住民税/配偶者控除額_老人控除対象配偶者.csv",
                         delimiter=",", skip_header=1, dtype="int64")[np.newaxis, 1:]


@cache
def 配偶者特別控除額表():
    """
    csvファイルから値を取得

    配偶者特別控除額表()[配偶者の所得区分, 納税者本人の所得区分] の形で参照可能
    """
    return np.genfromtxt(COUNTRY_DIR + "/assets/住民税/配偶者特別控除額.csv",
                  delimiter=",", skip_header=1, dtype="int64")[:, 1:]


class 住民税障害者控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税における障害者控除額"
    reference = "https://www.tax.metro.tokyo.lg.jp/kazei/kojin_ju.html#gaiyo_07"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    """

    def formula(対象人物, 対象期間, _parameters):
        # 自身が扶養に入っている、または同一生計配偶者である場合、納税者（世帯主）が控除を受ける
        扶養親族である = 対象人物("扶養親族である", 対象期間)
        同一生計配偶者である = 対象人物("同一生計配偶者である", 対象期間)
        被扶養者である = 扶養親族である + 同一生計配偶者である

        所得 = 対象人物("所得", 対象期間)
        所得降順 = 対象人物.get_rank(対象人物.世帯, -所得)
        # NOTE: 便宜上、被扶養者は所得が最も高い世帯員の扶養に入るとする
        所得が最も高い世帯員である = 所得降順 == 0

        控除対象額 = 対象人物("住民税障害者控除対象額", 対象期間)
        # NOTE: 異なる人物に対する値であるため、人物ではなく世帯ごとに集計（でないと「扶養者である」と要素がずれてしまい計算できない)
        被扶養者の合計控除額 = 対象人物.世帯.sum(被扶養者である * 控除対象額)

        # 最も所得が高い世帯員ではないが、一定以上の所得がある場合
        扶養に入っていない納税者である = np.logical_not(所得が最も高い世帯員である) * np.logical_not(被扶養者である)
        # 被扶養者は控除を受けない（扶養者が代わりに控除を受けるため）
        return 所得が最も高い世帯員である * (被扶養者の合計控除額 + 控除対象額) + 扶養に入っていない納税者である * 控除対象額


class 住民税障害者控除対象額(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税障害者控除額の対象額"
    reference = "https://www.tax.metro.tokyo.lg.jp/kazei/kojin_ju.html#gaiyo_07"
    documentation = """
    実際の控除額は同一生計配偶者、扶養親族が該当する場合にも加算される
    """

    def formula(対象人物, 対象期間, parameters):
        # 障害者控除額は対象人物ごとに算出される
        # https://www.city.hirakata.osaka.jp/kosodate/0000000544.html
        同居特別障害者控除対象 = 対象人物("同居特別障害者控除対象", 対象期間)
        # 重複して該当しないよう、同居特別障害者控除対象の場合を除外
        特別障害者控除対象 = 対象人物("特別障害者控除対象", 対象期間) * np.logical_not(同居特別障害者控除対象)
        障害者控除対象 = 対象人物("障害者控除対象", 対象期間)

        同居特別障害者控除額 = parameters(対象期間).住民税.同居特別障害者控除額
        特別障害者控除額 = parameters(対象期間).住民税.特別障害者控除額
        障害者控除額 = parameters(対象期間).住民税.障害者控除額

        return 同居特別障害者控除対象 * 同居特別障害者控除額 + 特別障害者控除対象 * 特別障害者控除額 + 障害者控除対象 * 障害者控除額


class 住民税ひとり親控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税におけるひとり親控除額"
    reference = "https://www.tax.metro.tokyo.lg.jp/kazei/kojin_ju.html#gaiyo_07"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    """

    def formula_2020_01_01(対象人物, 対象期間, parameters):
        所得 = 対象人物("所得", 対象期間)
        # 児童扶養手当の対象と異なり、父母の遺棄・DV等は考慮しない
        # (参考：児童扶養手当 https://www.city.hirakata.osaka.jp/0000026828.html)
        親である = 対象人物.has_role(世帯.親)
        子である = 対象人物.has_role(世帯.子)
        対象ひとり親 = (対象人物.世帯.sum(親である) == 1) * (対象人物.世帯.sum(子である) >= 1)
        ひとり親控除額 = parameters(対象期間).住民税.ひとり親控除額
        ひとり親控除_所得制限額 = parameters(対象期間).住民税.ひとり親控除_所得制限額

        return 親である * ひとり親控除額 * 対象ひとり親 * (所得 < ひとり親控除_所得制限額)


class 住民税寡婦控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税における寡婦控除額"
    reference = "https://www.tax.metro.tokyo.lg.jp/kazei/kojin_ju.html#gaiyo_07"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    """

    def formula_2020_01_01(対象人物, 対象期間, parameters):
        所得 = 対象人物("所得", 対象期間)
        寡婦 = 対象人物("寡婦", 対象期間)
        寡婦控除額 = parameters(対象期間).住民税.寡婦控除額
        寡婦控除_所得制限額 = parameters(対象期間).住民税.寡婦控除_所得制限額

        return 寡婦控除額 * 寡婦 * (所得 <= 寡婦控除_所得制限額)


class 住民税勤労学生控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税における勤労学生控除"
    reference = "https://www.tax.metro.tokyo.lg.jp/kazei/kojin_ju.html#gaiyo_07"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    """

    def formula(対象人物, 対象期間, parameters):
        所得 = 対象人物("所得", 対象期間)
        学生である = 対象人物("学生", 対象期間)
        勤労学生控除額 = parameters(対象期間).住民税.勤労学生控除額
        勤労学生_所得制限額 = parameters(対象期間).住民税.勤労学生_所得制限額
        所得条件 = (所得 > 0) * (所得 <= 勤労学生_所得制限額)

        return 所得条件 * 学生である * 勤労学生控除額


class 住民税配偶者控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税における配偶者控除"
    reference = "https://www.city.hiroshima.lg.jp/soshiki/26/202040.html"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    配偶者特別控除とは異なる。
    配偶者の所得が配偶者控除の所得制限を超えた場合でも、配偶者特別控除が適用される可能性がある。
    老人控除対象かどうかは以下を参照。
    https://www.nta.go.jp/taxes/shiraberu/taxanswer/shotoku/1182.htm
    """

    def formula(対象人物, 対象期間, _parameters):
        # 所得が高いほうが控除を受ける対象となる
        所得一覧 = 対象人物("所得", 対象期間)
        所得降順 = 対象人物.get_rank(対象人物.世帯, - 所得一覧, condition=対象人物.has_role(世帯.親))
        控除対象者である = (所得降順 == 0) * 対象人物.has_role(世帯.親)
        控除対象者の配偶者である = (所得降順 == 1) * 対象人物.has_role(世帯.親)
        控除対象者の所得 = 所得一覧 * 控除対象者である
        # NOTE: 異なる人物に対する値であるため、人物ではなく世帯ごとに集計（でないと「控除対象者の所得」と要素がずれてしまい計算できない)
        控除対象者の配偶者の所得 = 対象人物.世帯.sum(所得一覧 * 控除対象者の配偶者である)

        控除対象者の所得区分 = np.select(
            [控除対象者の所得 <= 9000000,
             (控除対象者の所得 > 9000000) * (控除対象者の所得 <= 9500000),
             (控除対象者の所得 > 9500000) * (控除対象者の所得 <= 10000000)],
            list(range(3)),
            -1).astype(int)

        控除対象者の配偶者の所得区分 = np.select(
            [控除対象者の配偶者の所得 <= 480000],
            [0],
            -1).astype(int)

        対象所得区分に該当する = (控除対象者の所得区分 != -1) * (控除対象者の配偶者の所得区分 != -1)  # 控除条件

        # NOTE: その年の12/31時点の年齢を参照
        # https://www.nta.go.jp/taxes/shiraberu/taxanswer/yogo/senmon.htm#word5
        該当年12月31日 = period(f"{対象期間.start.year}-12-31")
        該当年12月31日の年齢一覧 = 対象人物("年齢", 該当年12月31日)
        控除対象者の配偶者の年齢 = 該当年12月31日の年齢一覧 * 控除対象者の配偶者である
        # NOTE: 自分ではない人物についての計算のため、世帯で計算（でないと要素がずれてしまい計算できない）
        配偶者が老人控除対象である = 対象人物.世帯.sum(控除対象者の配偶者の年齢 >= 70)
        老人控除対象配偶者控除額 = 老人控除対象配偶者_配偶者控除額表()[控除対象者の配偶者の所得区分, 控除対象者の所得区分]

        通常配偶者控除額 = 配偶者控除額表()[控除対象者の配偶者の所得区分, 控除対象者の所得区分]

        配偶者控除額 = np.logical_not(配偶者が老人控除対象である) * 通常配偶者控除額 + 配偶者が老人控除対象である * 老人控除対象配偶者控除額

        配偶者がいる = 対象人物.世帯.sum(対象人物.has_role(世帯.親)) == 2  # 控除条件

        return 控除対象者である * 配偶者がいる * 対象所得区分に該当する * 配偶者控除額


class 住民税配偶者特別控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税における配偶者特別控除"
    reference = "https://www.city.hiroshima.lg.jp/soshiki/26/202040.html"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    """

    def formula(対象人物, 対象期間, _parameters):
        # 所得が高いほうが控除を受ける対象となる
        所得一覧 = 対象人物("所得", 対象期間)
        所得降順 = 対象人物.get_rank(対象人物.世帯, - 所得一覧, condition=対象人物.has_role(世帯.親))
        控除対象者である = (所得降順 == 0) * 対象人物.has_role(世帯.親)
        控除対象者の配偶者である = (所得降順 == 1) * 対象人物.has_role(世帯.親)
        控除対象者の所得 = 所得一覧 * 控除対象者である
        # NOTE: 異なる人物に対する値であるため、人物ではなく世帯ごとに集計（でないと「控除対象者の所得」と要素がずれてしまい計算できない)
        控除対象者の配偶者の所得 = 対象人物.世帯.sum(所得一覧 * 控除対象者の配偶者である)

        控除対象者の所得区分 = np.select(
            [控除対象者の所得 <= 9000000,
             (控除対象者の所得 > 9000000) * (控除対象者の所得 <= 9500000),
             (控除対象者の所得 > 9500000) * (控除対象者の所得 <= 10000000)],
            list(range(3)),
            -1).astype(int)

        控除対象者の配偶者の所得区分 = np.select(
            [(控除対象者の配偶者の所得 > 480000) * (控除対象者の配偶者の所得 <= 1000000),
             (控除対象者の配偶者の所得 > 1000000) * (控除対象者の配偶者の所得 <= 1050000),
             (控除対象者の配偶者の所得 > 1050000) * (控除対象者の配偶者の所得 <= 1100000),
             (控除対象者の配偶者の所得 > 1100000) * (控除対象者の配偶者の所得 <= 1150000),
             (控除対象者の配偶者の所得 > 1150000) * (控除対象者の配偶者の所得 <= 1200000),
             (控除対象者の配偶者の所得 > 1200000) * (控除対象者の配偶者の所得 <= 1250000),
             (控除対象者の配偶者の所得 > 1250000) * (控除対象者の配偶者の所得 <= 1300000),
             (控除対象者の配偶者の所得 > 1300000) * (控除対象者の配偶者の所得 <= 1330000)],
            list(range(8)),
            -1).astype(int)

        対象所得区分に該当する = (控除対象者の所得区分 != -1) * (控除対象者の配偶者の所得区分 != -1)  # 控除条件

        配偶者がいる = 対象人物.世帯.sum(対象人物.has_role(世帯.親)) == 2  # 控除条件

        return 控除対象者である * 配偶者がいる * 対象所得区分に該当する * 配偶者特別控除額表()[控除対象者の配偶者の所得区分, 控除対象者の所得区分]


class 住民税扶養控除(Variable):
    value_type = float
    entity = 人物
    definition_period = DAY
    label = "住民税における扶養控除"
    reference = "https://www.town.hinode.tokyo.jp/0000000519.html"
    documentation = """
    所得税における控除額とはことなるので注意（金額が異なるだけで条件は同じ）。
    OpenFiscaではクラス名をアプリ全体で一意にする必要があるため、先頭に「住民税」を追加。
    """

    def formula(対象人物, 対象期間, parameters):
        扶養親族である = 対象人物("扶養親族である", 対象期間)

        # NOTE: その年の12/31時点の年齢を参照
        # https://www.nta.go.jp/taxes/shiraberu/taxanswer/yogo/senmon.htm#word5
        該当年12月31日 = period(f"{対象期間.start.year}-12-31")
        年齢 = 対象人物("年齢", 該当年12月31日)

        控除対象扶養親族である = 扶養親族である * (年齢 >= 16)

        特定扶養親族である = 控除対象扶養親族である * (年齢 >= 19) * (年齢 < 23)
        老人扶養親族である = 控除対象扶養親族である * (年齢 >= 70)

        # NOTE: 入院中の親族は同居扱いだが老人ホーム等への入居は除く
        # TODO: 「同居していない親族」も世帯内で扱うようになったら同居老親かどうかの判定追加
        介護施設入所中 = 対象人物("介護施設入所中", 対象期間)
        同居している老人扶養親族である = 老人扶養親族である * np.logical_not(介護施設入所中)
        同居していない老人扶養親族である = 老人扶養親族である * 介護施設入所中

        # NOTE: np.selectのcondlistは最初に該当した条件で計算される
        扶養控除額 = np.select(
            [特定扶養親族である,
             同居している老人扶養親族である,
             同居していない老人扶養親族である,
             控除対象扶養親族である],
            [parameters(対象期間).住民税.扶養控除_特定扶養親族,
             parameters(対象期間).住民税.扶養控除_老人扶養親族_同居老親等,
             parameters(対象期間).住民税.扶養控除_老人扶養親族_同居老親等以外の者,
             parameters(対象期間).住民税.扶養控除_一般],
            0)

        所得 = 対象人物("所得", 対象期間)
        所得降順 = 対象人物.get_rank(対象人物.世帯, -所得)
        # NOTE: 便宜上所得が最も多い世帯員が扶養者であるとする
        扶養者である = 所得降順 == 0

        return 扶養者である * 対象人物.世帯.sum(扶養控除額)


class 住民税基礎控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "住民税における基礎控除"
    reference = "https://www.city.yokohama.lg.jp/kurashi/koseki-zei-hoken/zeikin/y-shizei/kojin-shiminzei-kenminzei/kaisei/R3zeiseikaisei.html"

    def formula(対象世帯, 対象期間, _parameters):
        世帯高所得 = 対象世帯("世帯高所得", 対象期間)

        return np.select(
            [世帯高所得 <= 24000000,
             (世帯高所得 > 24000000) * (世帯高所得 <= 24500000),
             (世帯高所得 > 24500000) * (世帯高所得 <= 25000000)],
            [430000,
             290000,
             150000],
            0)


class 控除後住民税世帯高所得(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "住民税計算において、各種控除が適用された後の世帯高所得額"
    reference = "https://www.town.hinode.tokyo.jp/0000000516.html"

    def formula(対象世帯, 対象期間, _parameters):
        # TODO: 社会保険料を追加
        世帯高所得 = 対象世帯("世帯高所得", 対象期間)
        配偶者控除一覧 = 対象世帯.members("住民税配偶者控除", 対象期間)
        配偶者控除 = 対象世帯.sum(配偶者控除一覧)
        配偶者特別控除一覧 = 対象世帯.members("住民税配偶者特別控除", 対象期間)
        配偶者特別控除 = 対象世帯.sum(配偶者特別控除一覧)
        扶養控除一覧 = 対象世帯.members("住民税扶養控除", 対象期間)
        扶養控除 = 対象世帯.sum(扶養控除一覧)
        障害者控除一覧 = 対象世帯.members("住民税障害者控除", 対象期間)
        障害者控除 = 対象世帯.sum(障害者控除一覧)
        ひとり親控除一覧 = 対象世帯.members("住民税ひとり親控除", 対象期間)
        ひとり親控除 = 対象世帯.sum(ひとり親控除一覧)
        寡婦控除一覧 = 対象世帯.members("住民税寡婦控除", 対象期間)
        寡婦控除 = 対象世帯.sum(寡婦控除一覧)
        勤労学生控除一覧 = 対象世帯.members("住民税勤労学生控除", 対象期間)
        勤労学生控除 = 対象世帯.sum(勤労学生控除一覧)
        基礎控除 = 対象世帯("住民税基礎控除", 対象期間)

        # 他の控除（雑損控除・医療費控除等）は定額でなく実費を元に算出するため除外する

        総控除額 = 配偶者控除 + 配偶者特別控除 + 扶養控除 + 障害者控除 + \
            ひとり親控除 + 寡婦控除 + 勤労学生控除 + 基礎控除

        # 負の数にならないよう、0円未満になった場合は0円に補正
        return np.clip(世帯高所得 - 総控除額, 0.0, None)


class 住民税非課税世帯(Variable):
    value_type = bool
    default_value = False
    entity = 世帯
    definition_period = DAY
    label = "住民税非課税世帯か否か（東京23区で所得割と均等割両方が非課税になる世帯）"
    reference = "https://financial-field.com/tax/entry-173575"

    # 市町村の級地により住民税均等割における非課税限度額が異なる
    # https://www.soumu.go.jp/main_content/000758656.pdf

    def formula(対象世帯, 対象期間, _parameters):
        世帯高所得 = 対象世帯("世帯高所得", 対象期間)
        世帯人数 = 対象世帯("世帯人数", 対象期間)
        居住級地区分1 = 対象世帯("居住級地区分1", 対象期間)

        級地区分倍率 = np.select([居住級地区分1 == 1, 居住級地区分1 == 2, 居住級地区分1 == 3],
                         [1, 0.9, 0.8],
            1)

        加算額 = np.select([世帯人数 == 1, 世帯人数 > 1],
                        [0, 210000 * 級地区分倍率],
                        0)

        return 世帯高所得 <= (350000 * 級地区分倍率 * 世帯人数 + 100000 + 加算額)


class 人的控除額の差(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "所得税と住民税の控除の差額"
    reference = "https://money-bu-jpx.com/news/article043882/"
    documentation = """
    差額計算には一部例外あり
    例外についての詳細は https://www.town.hinode.tokyo.jp/0000000519.html も参考になる
    """

    def formula(対象世帯, 対象期間, _parameters):
        障害者控除差額 = 対象世帯.sum(対象世帯.members("障害者控除", 対象期間) - 対象世帯.members("住民税障害者控除", 対象期間))
        寡婦控除一覧 = 対象世帯.members("寡婦控除", 対象期間)
        住民税寡婦控除一覧 = 対象世帯.members("住民税寡婦控除", 対象期間)
        寡婦控除差額 = 対象世帯.sum(寡婦控除一覧 - 住民税寡婦控除一覧)
        勤労学生控除一覧 = 対象世帯.members("勤労学生控除", 対象期間)
        住民税勤労学生控除一覧 = 対象世帯.members("住民税勤労学生控除", 対象期間)
        勤労学生控除差額 = 対象世帯.sum(勤労学生控除一覧 - 住民税勤労学生控除一覧)
        配偶者控除一覧 = 対象世帯.members("配偶者控除", 対象期間)
        住民税配偶者控除一覧 = 対象世帯.members("住民税配偶者控除", 対象期間)
        配偶者控除差額 = 対象世帯.sum(配偶者控除一覧 - 住民税配偶者控除一覧)
        配偶者特別控除一覧 = 対象世帯.members("配偶者特別控除", 対象期間)
        住民税配偶者特別控除一覧 = 対象世帯.members("住民税配偶者特別控除", 対象期間)
        配偶者特別控除差額 = 対象世帯.sum(配偶者特別控除一覧 - 住民税配偶者特別控除一覧)
        扶養控除差額 = 対象世帯.sum(対象世帯.members("扶養控除", 対象期間) - 対象世帯.members("住民税扶養控除", 対象期間))

        # NOTE: 以下は実際の差額とは異なる計算式を使用 https://www.town.hinode.tokyo.jp/0000000519.html
        世帯高所得 = 対象世帯("世帯高所得", 対象期間)
        基礎控除差額 = np.where(世帯高所得 <= 25000000, 50000, 0)

        通常ひとり親控除差額一覧 = 対象世帯.members("ひとり親控除", 対象期間)
        住民税通常ひとり親控除差額一覧 = 対象世帯.members("住民税ひとり親控除", 対象期間)
        通常ひとり親控除差額 = 対象世帯.sum(通常ひとり親控除差額一覧 - 住民税通常ひとり親控除差額一覧)
        # ひとり親（父）の場合のみ差額が異なる
        性別一覧 = 対象世帯.members("性別", 対象期間)
        親である = 対象世帯.has_role(世帯.親)
        父親がいる = 対象世帯.sum((性別一覧 == 性別パターン.男性) * 親である)
        ひとり親父世帯である = (対象世帯.nb_persons(世帯.親) == 1) * 父親がいる

        ひとり親控除差額 = np.logical_not(ひとり親父世帯である) * 通常ひとり親控除差額 + (通常ひとり親控除差額 > 0) * ひとり親父世帯である * 10000

        return 障害者控除差額 + 寡婦控除差額 + 勤労学生控除差額 + 配偶者控除差額 + 配偶者特別控除差額 + 扶養控除差額 + 基礎控除差額 + ひとり親控除差額


class 調整控除(Variable):
    value_type = float
    entity = 世帯
    definition_period = DAY
    label = "調整控除"
    reference = "https://money-bu-jpx.com/news/article043882/"
    documentation = """
    所得税と住民税の控除の差額（一部例外あり）
    例外についての詳細は https://www.town.hinode.tokyo.jp/0000000519.html も参考になる
    """

    def formula(対象世帯, 対象期間, parameters):
        人的控除額の差 = 対象世帯("人的控除額の差", 対象期間)
        # 個人住民税の課税所得金額に相当
        控除後住民税世帯高所得 = 対象世帯("控除後住民税世帯高所得", 対象期間)
        控除後住民税世帯高所得と人的控除額の差の小さい方 =\
            控除後住民税世帯高所得 * (控除後住民税世帯高所得 < 人的控除額の差) + 人的控除額の差 * (控除後住民税世帯高所得 >= 人的控除額の差)

        控除額 = np.select(
            [控除後住民税世帯高所得 <= 2000000,
             (控除後住民税世帯高所得 > 2000000) * (控除後住民税世帯高所得 < 25000000)],
            [控除後住民税世帯高所得と人的控除額の差の小さい方 * 0.05,
             (人的控除額の差 - (控除後住民税世帯高所得 - 2000000)) * 0.05],
            0)

        # 負の数にならないよう、0円未満になった場合は0円に補正
        return np.clip(控除額, 0.0, None)
